// =============================================================================
//  TRUFROM MD2/MD3-Viewer - 2003 by Sascha Willems
//  visit http://www.delphigl.de for more
//
//  !!! If you use this code in your projects, then please give me proper !!!
//  !!! credit and let me know                                            !!!
//
//  Credits and thanks to :
//
//   Jan Horn (R.I.P.) for the MD3-Loader and Textureloader (http://sulaco.co.za)
//   Tom Nuydens for the MD2-Loader (http://www.delphi3d.net)
//   James Green for the PKnight-MD2-Moodel (james@perilith.com)
//   ATI for making such a great graphicscard!
//
//  This is a viewer for the MD2 and MD3-Modelfileformat used in Quake2/Quake3
//  and many games based on those engines.
//  It enhances the lowpoly-models with TRUFORM and makes them look much better
//
//  Note that some models may not work correctly with TRUFORM...
//
//  For more see :
//   http://www.ati.com/developer/sdk/RadeonSDK/Html/Info/ati_pn_triangles.txt
//   http://www.ati.com/developer/sdk/RadeonSDK/Html/Samples/OpenGL/PNTriangles.html
//
//  Graphicscards supporting this extension :
//
//    Radeon8500 Series
//    Radeon9500/9600/9700/9800
//
//  If you have one of these, and it won't tesselate the mesh, then you'll have
//  to enable TRUFORM in the opengl-tab of your driver.This seems to be disabled
//  by default on newer drivers/boards...
// =============================================================================
//  Revisionhistory :
//   Version 1.0
//    - First pulic release
// =============================================================================

unit TruformModelView1;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  ExtCtrls,
  AppEvnts,
  StdCtrls,
  jpeg,

  OpenGL12,
  glExtensions,
  Textures,
  LoadMD2,
  LoadMD3;

type
  TForm1 = class(TForm)
    ApplicationEvents: TApplicationEvents;
    FrameTimer: TTimer;
    StatPanel: TPanel;
    glPanel: TPanel;
    OpenDialog1: TOpenDialog;
    OpenDialog2: TOpenDialog;
    OpenDialog3: TOpenDialog;
    Panel1: TPanel;
    AnimList: TListBox;
    OpenMD3Button: TButton;
    OpenMD2Button: TButton;
    UpdateInfoLabel: TTimer;

    TRUFORMButton: TButton;
    TRUFORMLevelButton: TButton;
    WireframeButton: TButton;
    InfoLabel: TLabel;
    TextureButton: TButton;
    ATILogo: TImage;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
    procedure FrameTimerTimer(Sender: TObject);
    procedure AnimListClick(Sender: TObject);
    procedure TRUFORMLevelButtonClick(Sender: TObject);
    procedure TRUFORMButtonClick(Sender: TObject);
    procedure WireframeButtonClick(Sender: TObject);
    procedure OpenMD3ButtonClick(Sender: TObject);
    procedure OpenMD2ButtonClick(Sender: TObject);
    procedure GLPanelResize(Sender: TObject);
    procedure UpdateInfoLabelTimer(Sender: TObject);
    procedure TextureButtonClick(Sender: TObject);
  private
    { Private-Deklarationen }
  public
   RC                 : HGLRC;
   DC                 : HDC;
   UseTRUFORM         : Boolean;
   UseTexture         : Boolean;
   WireFrame          : Boolean;
   PNTesselationLevel : Integer;
   MaxPNTessLevel     : Integer;
   Frames             : Integer;
   FPS                : Integer;
   procedure Draw;
  end;
 TModelType = (mtMD2, mtMD3);

var
 Form1           : TForm1;
 ModelType       : TModelType;
 DoIdle          : Boolean;
 Q2Tex           : TGLUInt;
 Q2AnimIndex     : Integer;
 Q2Anim          : Single;
 PerfCounterFreq : Int64;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
var
 DummyPal : HPalette;
begin
// Get the frequency of the performancecounter for timebased-animation
QueryPerformanceFrequency(PerfCounterFreq);
// Get a DC and create a rendercontext
DC := GetDC(GLPanel.Handle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 32, 0, 0, 0, 0, DummyPal);
ActivateRenderingContext(DC, RC);
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LEQUAL);
// Tell OpenGL to normalize our Normals for correct lighting
glEnable(GL_NORMALIZE);
glClearColor(0.6, 0.6, 0.6, 1);
UseTRUFORM         := True;
UseTexture         := True;
PNTesselationLevel := 1;
Q2Anim             := 0;
Q2AnimIndex        := 0;
// Load Functionpointers
glPNTrianglesiATI := wglGetProcAddress('glPNTrianglesiATI');
// Check for TRUFORM-Support
if not Assigned(glPNTrianglesiATI) then
 begin
 ShowMessage('GL_PN_TRIANGLES_ATI not supported by Hardware!');
 Application.Terminate;
 end;
// Get maximum PN-Tesselationlevel
glGetIntegerV(GL_MAX_PN_TRIANGLES_TESSELATION_LEVEL_ATI, @MaxPNTessLevel);
// Load MD2-Model
ModelType := mtMD2;
ChDir(ExtractFilePath(Application.ExeName));
AnimList.Items.LoadFromFile('anims_md2.txt');
LoadModel('pknight\tris.md2');
ChDir(ExtractFilePath(Application.ExeName));
glEnable(GL_TEXTURE_2D);
LoadTexture('pknight\evil.jpg', Q2Tex, False);
// Start rendering
DoIdle := True;
OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
OpenDialog2.InitialDir := ExtractFilePath(Application.ExeName);
OpenDialog3.InitialDir := ExtractFilePath(Application.ExeName);
end;

procedure TForm1.Draw;
const
 Light0 : array[0..3] of TGLFloat = (1,-1,1,0);
 Light1 : array[0..3] of TGLFloat = (1,1,1,0);
 Light2 : array[0..3] of TGLFloat = (-1,1,1,0);
 Color1 : array[0..3] of TGLFloat = (0.6,0.3,0.6,1);
 Color2 : array[0..3] of TGLFloat = (0.3,0.6,0.6,1);
 Spec   : array[0..3] of TGLFloat = (1, 1, 1, 1);
 White  : array[0..3] of TGLFloat = (0.8, 0.8, 0.8, 1);
 Blue   : array[0..3] of TGLFloat = (0.1, 0.1, 1, 1);
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(60, glPanel.ClientWidth/glPanel.ClientHeight, 1, 512);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);

if WireFrame then
 glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
else
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

glLightfv(GL_LIGHT0, GL_POSITION, @Light0);
glLightfv(GL_LIGHT1, GL_POSITION, @Light1);
glLightfv(GL_LIGHT2, GL_POSITION, @Light2);
glLightfv(GL_LIGHT1, GL_DIFFUSE, @Color1);
glLightfv(GL_LIGHT1, GL_SPECULAR, @Color1);
glLightfv(GL_LIGHT2, GL_DIFFUSE, @Color2);
glLightfv(GL_LIGHT2, GL_SPECULAR, @Color2);
glEnable(GL_LIGHTING);
glEnable(GL_LIGHT0);
glEnable(GL_LIGHT1);
glEnable(GL_LIGHT2);

glColor3f(1,1,1);
glMaterialfv(GL_FRONT, GL_SPECULAR, @Spec);
glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @White);
glMaterialf(GL_FRONT, GL_SHININESS, 32);

glTranslatef(0, 0, -100);

if UseTRUFORM then
 begin
 glEnable(GL_PN_TRIANGLES_ATI);
 glPNTrianglesiATI(GL_PN_TRIANGLES_TESSELATION_LEVEL_ATI, PNTesselationLevel);
 end;

case ModelType of
 mtMD2 : begin
         if UseTexture then
          begin
          glEnable(GL_TEXTURE_2D);
          glBindTexture(GL_TEXTURE_2D, Q2Tex);
          // flip texture upside down cause texcoords in md2 are upside down
          glMatrixMode(GL_TEXTURE);
          glLoadIdentity;
          glScalef(1,-1,1);
          glMatrixMode(GL_MODELVIEW);
          end
         else
          glDisable(GL_TEXTURE_2D);
         glPushMatrix;
          glScalef(1.6, 1.6, 1.6);
          glRotatef(60, 0, -1, 0);
          glRotatef(90, -1, 0, 0);
          // Enable Texturecoord-, Normal- and Vertexarrays
          glEnableClientState(GL_TEXTURE_COORD_ARRAY);
          glEnableClientState(GL_NORMAL_ARRAY);
          glEnableClientState(GL_VERTEX_ARRAY);
          // Tell OpenGL where to find the vertices, normals and Texcoords
          glTexCoordPointer(2, GL_FLOAT, SizeOf(LoadMD2.TVertex), @VertexArrays[Round(Q2Anim)][0].Tex0);
          glNormalPointer(GL_FLOAT, SizeOf(LoadMD2.TVertex), @VertexArrays[Round(Q2Anim)][0].Nrml);
          glVertexPointer(3, GL_FLOAT, SizeOf(LoadMD2.TVertex), @VertexArrays[Round(Q2Anim)][0].OPos);
          // Draw the vertexarray
          glDrawArrays(GL_TRIANGLES, 0, Length(VertexArrays[Round(Q2Anim)]));
         glPopMatrix;
         glMatrixMode(GL_TEXTURE);
         glLoadIdentity;
         glMatrixMode(GL_MODELVIEW);
         end;
 mtMD3 : begin
         if UseTexture then
          glEnable(GL_TEXTURE_2D)
         else
          glDisable(GL_TEXTURE_2D);
         glPushMatrix;
          glTranslatef(0,-6,0);
          glRotatef(90, -1, 0, 0);
          glRotatef(90, 0, 0, -1);
          glScalef(1.3, 1.3, 1.3);
          // Draw MD3-Model
          Q3Model.Draw(Q3Model.AnimPhase);
         glPopMatrix;
         end;
end;
glDisable(GL_PN_TRIANGLES_ATI);

SwapBuffers(DC);
inc(Frames);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
if (DC<>0) and (RC<>0) then
 begin
 DeActivateRenderingContext;
 wglDeleteContext(RC);
 ReleaseDC(GLPanel.Handle, DC);
 end;
end;

procedure TForm1.ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
var
 QPCStartCount, QPCEndCount : Int64;
begin
if not DoIdle then
 exit;
QueryPerformanceCounter(QPCStartCount);
Draw;
QueryPerformanceCounter(QPCEndCount);
case ModelType of
 mtMD2 : begin
         // Calculate animationtime for md2
         Q2Anim := Q2Anim+0.1*(QPCEndCount-QPCStartCount)/PerfCounterFreq*100;
         if Q2Anim >= MD2_ANIM_FRAME_NUMBERS[Q2AnimIndex, 0]+MD2_ANIM_FRAME_NUMBERS[Q2AnimIndex, 1]-1  then
          Q2Anim := MD2_ANIM_FRAME_NUMBERS[Q2AnimIndex, 0];
         end;
 mtMD3 : begin
         // Calculate animationtime for md3
         Q3Model.AnimPhase := Q3Model.AnimPhase+0.01*(QPCEndCount-QPCStartCount)/PerfCounterFreq*100;
         end;
end;
Done := False;
end;

procedure TForm1.FrameTimerTimer(Sender: TObject);
begin
FPS    := Frames;
Frames := 0;
end;

procedure TForm1.AnimListClick(Sender: TObject);
begin
if AnimList.ItemIndex >= 0 then
 case ModelType of
  mtMD3 : Q3Model.SetAnim(AnimList.ItemIndex);
  mtMD2 : begin
          Q2AnimIndex := AnimList.ItemIndex;
          Q2Anim      := MD2_ANIM_FRAME_NUMBERS[Q2AnimIndex, 0];
          end;
 end;
end;

procedure TForm1.GLPanelResize(Sender: TObject);
begin
if (DC<>0) and (RC<>0) then
 begin
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 gluPerspective(60, glPanel.ClientWidth/glPanel.ClientHeight, 1, 512);
 glViewPort(0,0,glPanel.ClientWidth, glPanel.ClientHeight);
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity;
 end;
end;

procedure TForm1.UpdateInfoLabelTimer(Sender: TObject);
begin
InfoLabel.Caption := 'PN-Level = '+IntToStr(PNTesselationLevel)+'/'+IntToStr(MaxPNTessLevel)+'   '+'FPS = '+IntToStr(FPS);
end;

procedure TForm1.TRUFORMLevelButtonClick(Sender: TObject);
begin
if PNTesselationLevel < MaxPNTessLevel then
 inc(PNTesselationLevel)
else
 PNTesselationLevel := 1;
end;

procedure TForm1.TRUFORMButtonClick(Sender: TObject);
begin
UseTRUFORM := not UseTRUFORM;
case UseTRUFORM of
 False : TRUFORMButton.Caption := 'turn TRUFORM on';
 True  : TRUFORMButton.Caption := 'turn TRUFORM off';
end;
end;

procedure TForm1.WireframeButtonClick(Sender: TObject);
begin
WireFrame := not WireFrame;
end;

procedure TForm1.OpenMD3ButtonClick(Sender: TObject);
begin
if OpenDialog1.Execute then
 Q3Model.LoadPlayer(ExtractFilePath(OpenDialog1.FileName)+'\', 'default');
ChDir(ExtractFilePath(Application.ExeName));
AnimList.Items.LoadFromFile('anims_md3.txt');
ModelType := mtMD3;
end;

procedure TForm1.OpenMD2ButtonClick(Sender: TObject);
begin
if not OpenDialog2.Execute then
 exit;
if not OpenDialog3.Execute then
 exit;
LoadModel(OpenDialog2.FileName);
LoadTexture(OpenDialog3.FileName, Q2Tex, False);
Q2Anim := 0;
ChDir(ExtractFilePath(Application.ExeName));
AnimList.Items.LoadFromFile('anims_md2.txt');
ModelType := mtMD2;
end;

procedure TForm1.TextureButtonClick(Sender: TObject);
begin
useTexture := not UseTexture;
case UseTexture of
 False : TextureButton.Caption := 'turn TEXTURE on';
 True  : TextureButton.Caption := 'turn TEXTURE off';
end;
end;

end.
